package com.example.calculator

import android.app.Activity
import android.util.Log
import android.view.View
import android.widget.Button
import android.widget.TextView
import net.objecthunter.exp4j.ExpressionBuilder
import java.util.*
import kotlin.collections.ArrayList

/*
 * A Calculator uses PEMDAS and parses input as a solvable equation.
 * Supports decimal calculations.
 * Displays an intermediate result on top in red small letters and the final result in black.
 * TODO: Partially usable with TalkBack. How can you improve the user experience?
*/
class Calculator(_activity: Activity) {

    private var mActivity: Activity = _activity
    private var mCalculatorModel: CalculatorModel = CalculatorModel()

    private var mResultView: TextView? = null // Contains the final result of calculation
    private var mIntermediateView: TextView? = null // Contains the intermediate result of input
    private var mButtonViews: ArrayList<Button>? = null // Calculator buttons

    init { // constructor, initializes references to all views from activity
        mButtonViews = (mActivity.findViewById<View>(R.id.parent_container)).touchables
                as ArrayList<Button>
        mResultView = mActivity.findViewById(R.id.result)
        mIntermediateView = mActivity.findViewById(R.id.intermediate_result)

        for (btn in mButtonViews!!) { // handles click responses for each button
            btn.setOnClickListener {
                if (btn.text.contains("=")) {
                    mCalculatorModel.mFinalResult = mCalculatorModel.mIntermediateResult
                    calculate(mCalculatorModel.mFinalResult)
                } else {
                    mCalculatorModel.mIntermediateResult += btn.text
                    updateViews()
                }
            }
        }

        mActivity.findViewById<Button>(R.id.cancel).setOnClickListener { reset() }

        mActivity.findViewById<Button>(R.id.back).setOnClickListener {
            if (mCalculatorModel.mIntermediateResult.isNotEmpty()) {
                mCalculatorModel.mIntermediateResult =
                        mCalculatorModel.mIntermediateResult.substring(0,
                        mCalculatorModel.mIntermediateResult.lastIndex)
                updateViews()
            }
        }
    }

    // Calculates and outputs a result for a given equation
    private fun calculate(stringResult: String) {
        try {
            val e = ExpressionBuilder(stringResult).build()
            val result = e.evaluate()
            mCalculatorModel.mFinalResult = result.toString()
            mCalculatorModel.mIntermediateResult = mCalculatorModel.mFinalResult
            updateViews()
        } catch (e: IllegalArgumentException) {
            handleError(stringResult) // bad user input
        } catch (e: EmptyStackException) {
            handleError(stringResult)
        }
    }

    // Resets state of calculator model
    private fun reset() {
        mCalculatorModel.mIntermediateResult = ""
        mCalculatorModel.mFinalResult = "0.0"
        updateViews()
    }

    // Catches errors in expression parsing
    private fun handleError(causeOfError: String) {
        Log.i("DEBUG", "Invalid input: $causeOfError")
        mIntermediateView?.text = ""
        mIntermediateView?.announceForAccessibility(mActivity.getString(R.string.error_msg))
        reset()
    }

    // Updates the appearance of the result
    private fun updateViews() {
        mResultView?.text = mCalculatorModel.mFinalResult
        mIntermediateView?.text = mCalculatorModel.mIntermediateResult
        mIntermediateView?.announceForAccessibility(
                mCalculatorModel.mIntermediateResult) // announce result
    }
}